from __future__ import annotations

import logging
from copy import deepcopy
from typing import NewType, Optional, Union

import numpy as np
from pyaro import list_timeseries_engines, open_timeseries
from pyaro.timeseries import Data, Reader, Station
from pyaro.timeseries.Wrappers import VariableNameChangingReader

from pyaerocom.io.pyaro.pyaro_config import PyaroConfig
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.tstype import TsType
from pyaerocom.ungriddeddata import UngriddedData

logger = logging.getLogger(__name__)


Metadata = NewType("Metadata", dict[str, dict[str, Union[str, list[str]]]])


class ReadPyaro(ReadUngriddedBase):
    __version__ = "1.0.1"

    SUPPORTED_DATASETS = list(list_timeseries_engines().keys())

    def __init__(self, config: PyaroConfig) -> None:
        self.config: PyaroConfig = config

        self._check_id()

        self.converter = PyaroToUngriddedData(self.config)
        self.reader = self.converter.reader
        self._data_dir = self.config.filename_or_obj_or_url
        self._data_name = self.config.name
        self._data_id = self.config.name

    """
    Definition of abstract methods from ReadUngriddedBase
    """

    @property
    def DATA_ID(self):
        return self._data_name

    @property
    def PROVIDES_VARIABLES(self):
        """
        return self.reader.get_variables()
        """
        return self.reader.variables()

    @property
    def DEFAULT_VARS(self):
        return self.PROVIDES_VARIABLES

    @property
    def TS_TYPE(self):
        """
        To be provided by the reader or engine
        """
        return "undefined"

    @property
    def _FILEMASK(self):
        return self.config.filename_or_obj_or_url

    @staticmethod
    def get_pyaro_readers():
        return list_timeseries_engines()

    def read(self, vars_to_retrieve=None, files=..., first_file=None, last_file=None):
        return self.converter.read(vars_to_retrieve=vars_to_retrieve)

    def read_file(self, filename, vars_to_retrieve=None):
        return self.converter.read(vars_to_retrieve)

    def _check_id(self):
        avail_readers = list_timeseries_engines()
        if not self.config.data_id in avail_readers:
            logger.warning(
                f"Could not find {self.config.data_id} in list of available Pyaro readers: {avail_readers}"
            )


class PyaroToUngriddedData:
    _METADATAKEYINDEX = 0
    _TIMEINDEX = 1
    _LATINDEX = 2
    _LONINDEX = 3
    _ALTITUDEINDEX = 4  # altitude of measurement device
    _VARINDEX = 5
    _DATAINDEX = 6
    _DATAHEIGHTINDEX = 7
    _DATAERRINDEX = 8  # col where errors can be stored
    _DATAFLAGINDEX = 9  # can be used to store flags
    _STOPTIMEINDEX = 10  # can be used to store stop time of acq.
    _TRASHINDEX = 11  # index where invalid data can be moved to (e.g. when outliers are removed)

    def __init__(self, config: PyaroConfig) -> None:
        self.data: UngriddedData = UngriddedData()
        self.config = config
        self.reader: Reader = self._open_reader()

    def _open_reader(self) -> Reader:
        data_id = self.config.data_id

        if self.config.name_map is None:
            return open_timeseries(
                data_id, self.config.filename_or_obj_or_url, filters=self.config.filters
            )
        else:
            return VariableNameChangingReader(
                open_timeseries(
                    data_id, self.config.filename_or_obj_or_url, filters=self.config.filters
                ),
                self.config.name_map,
            )

    def _convert_to_ungriddeddata(self, pyaro_data: dict[str, Data]) -> UngriddedData:
        stations = self.get_stations()

        var_size = {var: len(pyaro_data[var]) for var in pyaro_data}
        vars = list(pyaro_data.keys())
        total_size = sum(list(var_size.values()))
        units = {var: {"units": pyaro_data[var]._units} for var in pyaro_data}
        ts_types: dict[str, Optional[TsType]] = {k: None for k in stations}

        # Object necessary for ungriddeddata
        var_idx = {var: i for i, var in enumerate(vars)}
        metadata = self._make_ungridded_metadata(stations=stations, var_idx=var_idx, units=units)
        meta_idx = {s: {v: [] for v in vars} for s in metadata}
        data_array = np.zeros([total_size, 12])

        # Helper objects
        station_idx = {metadata[idx]["station_name"]: idx for idx in metadata}

        idx = 0
        for var, var_data in pyaro_data.items():
            size = var_size[var]
            for i in range(
                1, size
            ):  # The 1 start is a temp fix for the empty first row of the current Data implementation from pyaro
                data_line = var_data[i]
                current_station = data_line["stations"]

                # Fills data array
                ungriddeddata_line = self._pyaro_dataline_to_ungriddeddata_dataline(
                    data_line, idx, var_idx[var]
                )

                # Finds the ts_type of the stations. Raises error of same station has different types
                start, stop = data_line["start_times"], data_line["end_times"]
                ts_type = self._calculate_ts_type(start, stop)
                if ts_types[current_station] is None:
                    ts_types[current_station] = ts_type
                elif ts_types[current_station] != ts_type:
                    msg = f"TS type {ts_type} of station {current_station} is different from already found value {ts_types[current_station]}"
                    logger.error(msg)
                    raise ValueError(msg)

                data_array[idx, :] = ungriddeddata_line

                #  Fills meta_idx
                meta_idx[station_idx[current_station]][var].append(idx)

                idx += 1

        new_meta_idx = {}
        for station_id in meta_idx:
            new_meta_idx[station_id] = {}
            for var_id in meta_idx[station_id]:
                new_meta_idx[station_id][var_id] = np.array(meta_idx[station_id][var_id])

        self.data._data = data_array
        self.data.meta_idx = new_meta_idx
        self.data.metadata = self._add_ts_type_to_metadata(metadata, ts_types)
        self.data.var_idx = var_idx

        return self.data

    def _get_metadata_from_pyaro(self, station: Station) -> list[dict[str, str]]:
        metadata = dict(
            instrument_name=None,
            data_revision="n/d",
            PI=None,
            filename=None,
            country_code=station["country"],
            data_level=None,
            revision_date=None,
            website=None,
            data_product=None,
            data_version=None,
            framework=None,
            instr_vert_loc=None,
            stat_merge_pref_attr=None,
        )
        for key in metadata:
            if key in station.keys():
                metadata[key] = station[key]

        return metadata

    def _make_ungridded_metadata(
        self, stations: dict[str, Station], var_idx: dict[str, int], units: dict[str, str]
    ) -> Metadata:
        idx = 0
        metadata = {}
        for name, station in stations.items():
            metadata[idx] = dict(
                data_id=self.config.name,
                variables=list(self.get_variables()),
                var_info=units,
                latitude=station["latitude"],
                longitude=station["longitude"],
                altitude=station["altitude"],
                station_name=station["long_name"],
                station_id=name,
                country=station["country"],
                ts_type="undefined",  # TEMP: Changes dynamically below
            )

            metadata[idx].update(self._get_metadata_from_pyaro(station))
            idx += 1

        return Metadata(metadata)

    def _pyaro_dataline_to_ungriddeddata_dataline(
        self, data: np.void, idx: int, var_idx: int
    ) -> np.ndarray:
        new_data = np.zeros(12)
        new_data[self._METADATAKEYINDEX] = idx
        new_data[self._TIMEINDEX] = data["start_times"]
        new_data[self._LATINDEX] = data["latitudes"]
        new_data[self._LONINDEX] = data["longitudes"]
        new_data[self._ALTITUDEINDEX] = data["altitudes"]
        new_data[self._VARINDEX] = var_idx
        new_data[self._DATAINDEX] = data["values"]
        new_data[self._DATAHEIGHTINDEX] = np.nan
        new_data[self._DATAERRINDEX] = data["standard_deviations"]
        new_data[self._DATAFLAGINDEX] = data["flags"]
        new_data[self._STOPTIMEINDEX] = data["end_times"]
        new_data[self._TRASHINDEX] = np.nan

        return new_data

    def _calculate_ts_type(self, start: np.datetime64, stop: np.datetime64) -> TsType:
        seconds = (stop - start).astype("timedelta64[s]").astype(np.int32)
        if seconds == 0:
            ts_type = TsType("hourly")
        else:
            ts_type = TsType.from_total_seconds(seconds)

        return ts_type

    def _add_ts_type_to_metadata(
        self, metadata: Metadata, ts_types: dict[str, Optional[TsType]]
    ) -> Metadata:
        new_metadata: Metadata = deepcopy(metadata)
        for idx in new_metadata:
            station_name = new_metadata[idx]["station_name"]
            ts_type = str(ts_types[station_name])
            new_metadata[idx]["ts_type"] = ts_type if ts_type is not None else "undefined"
        return new_metadata

    def get_variables(self) -> list[str]:
        return self.reader.variables()

    def get_stations(self) -> dict[str, Station]:
        return self.reader.stations()()

    def read(self, vars_to_retrieve=None) -> UngriddedData:
        allowed_vars = self.get_variables()
        if vars_to_retrieve is None:
            vars_to_retrieve = allowed_vars
        else:
            if isinstance(vars_to_retrieve, str):
                vars_to_retrieve = [vars_to_retrieve]

        data = {}
        for var in vars_to_retrieve:
            if var not in allowed_vars:
                logger.warning(
                    f"Variable {var} not in list over allowed variabes for {self.config.data_id}: {allowed_vars}"
                )
                continue

            data[var] = self.reader.data(varname=var)

        return self._convert_to_ungriddeddata(data)



from __future__ import annotations

import json
from pathlib import Path

import numpy as np
from pytest import mark, param

from pyaerocom.aeroval.coldatatojson_helpers import (
    _add_heatmap_entry_json,
    _init_stats_dummy,
    _prepare_aerocom_regions_json,
    _prepare_country_regions,
    _prepare_default_regions_json,
    _prepare_htap_regions_json,
    _prepare_regions_json_helper,
    _write_diurnal_week_stationdata_json,
    _write_site_data,
    _write_stationdata_json,
    get_stationfile_name,
)
from pyaerocom.region import get_all_default_region_ids
from pyaerocom.region_defs import (
    HTAP_REGIONS,
    HTAP_REGIONS_DEFAULT,
    OLD_AEROCOM_REGIONS,
    OTHER_REGIONS,
)


def test__write_stationdata_json(tmp_path: Path):
    data = dict(station_name="stat1", obs_name="obs1", var_name_web="var1", vert_code="Column")
    path: Path = tmp_path / get_stationfile_name(**data)
    assert not path.exists()

    data["model_name"] = "model1"
    _write_stationdata_json(data, path.parent)
    assert path.exists()

    data["model_name"] = "model2"
    _write_stationdata_json(data, path.parent)
    assert path.exists()

    data = json.loads(path.read_text())
    assert data["model1"].pop("model_name") == "model1"
    assert data["model2"].pop("model_name") == "model2"
    assert data["model1"] == data["model2"]


def test__write_site_data(tmp_path: Path):
    data = [
        dict(
            model_name=f"model{n}",
            station_name=f"stat{n}",
            obs_name=f"obs{n}",
            var_name_web=f"var{n}",
            vert_code="Column",
        )
        for n in range(3)
    ]
    assert not list(tmp_path.glob("*.json"))
    _write_site_data(data, str(tmp_path))
    assert len(list(tmp_path.glob("*.json"))) == len(data)


def test__write_diurnal_week_stationdata_json(tmp_path: Path):
    data = dict(station_name="stat1", obs_name="obs1", var_name_web="var1", vert_code="Column")
    dirs = {"ts/diurnal": tmp_path}
    path: Path = tmp_path / get_stationfile_name(**data)
    assert not path.exists()

    data["model_name"] = "model1"
    _write_diurnal_week_stationdata_json(data, dirs)
    assert path.exists()

    data["model_name"] = "model2"
    _write_diurnal_week_stationdata_json(data, dirs)
    assert path.exists()

    data = json.loads(path.read_text())
    assert data["model1"].pop("model_name") == "model1"
    assert data["model2"].pop("model_name") == "model2"
    assert data["model1"] == data["model2"]


def test__add_heatmap_entry_json(tmp_path: Path):
    path = tmp_path / "add_entry.json"
    assert not path.exists()

    result = dict(value=None)
    obs, obs_var, vert, model, model_var = "obs", "obs_var", "Column", "model", "model_var"
    _add_heatmap_entry_json(path, result, obs, obs_var, vert, model, model_var)
    assert path.exists()

    data = json.loads(path.read_text())
    assert data[obs_var][obs][vert][model][model_var] == result

    result.update(value=42.0)
    _add_heatmap_entry_json(path, result, obs, obs_var, vert, model, model_var)

    data = json.loads(path.read_text())
    assert data[obs_var][obs][vert][model][model_var] == result


def test__init_stats_dummy():
    dummy = _init_stats_dummy()
    assert dummy
    assert all(np.isnan(value) for value in dummy.values())


@mark.parametrize(
    "region_ids",
    [
        param(OLD_AEROCOM_REGIONS, id="old AeroCom"),
        param(HTAP_REGIONS_DEFAULT, id="HTAP default"),
        param(HTAP_REGIONS, id="HTAP"),
        param(OTHER_REGIONS, id="other"),
    ],
)
def test__prepare_regions_json_helper(region_ids: list[str]):
    helper = _prepare_regions_json_helper(region_ids)
    assert len(helper) == 2
    assert len(helper[0]) == len(helper[1]) == len(region_ids)

    for region_name, borders in helper[0].items():
        assert region_name in helper[1]
        assert type(borders) == dict
        assert "minLat" in borders and -90 <= borders["minLat"] <= 90
        assert "maxLat" in borders and -90 <= borders["maxLat"] <= 90
        assert "minLon" in borders and -180 <= borders["minLon"] <= 180
        assert "maxLon" in borders and -180 <= borders["maxLon"] <= 180
        assert borders["minLat"] < borders["maxLat"]

    for region_name, region in helper[1].items():
        assert region_name in helper[0]
        assert region.region_id in region_ids
        assert region.name == region_name


def test__prepare_default_regions_json():
    borders, regions = _prepare_default_regions_json()
    region_ids = get_all_default_region_ids()
    assert bool(borders) and bool(regions) and bool(region_ids)
    assert list(borders) == list(regions)
    assert [reg.region_id for reg in regions.values()] == region_ids


def test__prepare_aerocom_regions_json():
    borders, regions = _prepare_aerocom_regions_json()
    region_ids = OLD_AEROCOM_REGIONS
    assert bool(borders) and bool(regions) and bool(region_ids)
    assert list(borders) == list(regions)
    assert [reg.region_id for reg in regions.values()] == region_ids


def test__prepare_htap_regions_json():
    borders, regions = _prepare_htap_regions_json()
    region_ids = HTAP_REGIONS_DEFAULT
    assert bool(borders) and bool(regions) and bool(region_ids)
    assert list(borders) == list(regions)
    assert [reg.region_id for reg in regions.values()] == region_ids


@mark.parametrize(
    "region_ids",
    [
        param(OLD_AEROCOM_REGIONS, id="old AeroCom"),
        param(HTAP_REGIONS_DEFAULT, id="HTAP default"),
        param(HTAP_REGIONS, id="HTAP"),
        param(OTHER_REGIONS, id="other"),
    ],
)
def test__prepare_country_regions(region_ids: list[str]):
    regions = _prepare_country_regions(region_ids)
    assert type(regions) is dict
    assert bool(regions)
    assert [reg.region_id for reg in regions.values()] == region_ids



from __future__ import annotations

from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from pyaerocom.io import PyaroConfig, ReadPyaro


def make_csv_test_file(tmp_path: Path) -> Path:
    d = tmp_path / "pyaro"
    file = d / "test_data.csv"

    if not d.exists():
        d.mkdir()
    if file.exists():
        return file

    start = pd.to_datetime("01.01.2015", dayfirst=True)
    end = pd.to_datetime("31.12.2015", dayfirst=True)
    dates = pd.date_range(start, end, freq="D")
    stations = ["NO0002", "GB0881"]
    coords = [(58, 8), (60, -1)]
    species = ["NOx", "SOx"]

    with open(file, "w") as f:
        for s in species:
            for i, station in enumerate(stations):
                for date in dates:
                    f.write(
                        f"{s}, {station}, {coords[i][1]}, {coords[i][0]}, {np.random.normal(10, 5)}, Gg, {date}, {date+pd.Timedelta('1D')} \n"
                    )

    return file


def testconfig(tmp_path: Path) -> PyaroConfig:
    data_id = "csv_timeseries"

    config1 = PyaroConfig(
        name="test",
        data_id=data_id,
        filename_or_obj_or_url=str(make_csv_test_file(tmp_path)),
        filters={},
        name_map={"SOx": "concso4"},
    )

    config2 = PyaroConfig(
        name="test2",
        data_id=data_id,
        filename_or_obj_or_url=str(make_csv_test_file(tmp_path)),
        filters={},
        name_map={"SOx": "concso4"},
    )
    return [config1, config2]


@pytest.fixture
def pyaro_test_data_file(tmp_path) -> Path:
    return make_csv_test_file(tmp_path)


@pytest.fixture
def pyaro_testconfig(tmp_path) -> PyaroConfig:
    return testconfig(tmp_path=tmp_path)


@pytest.fixture
def pyaro_testdata(tmp_path) -> ReadPyaro:
    config = testconfig(tmp_path=tmp_path)[0]
    rp = ReadPyaro(config=config)

    return rp



from __future__ import annotations

import logging
import os
from datetime import datetime
from pathlib import Path
from typing import Tuple

import numpy as np
from tqdm import tqdm

from pyaerocom import const
from pyaerocom._lowlevel_helpers import BrowseDict
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.plugins.icpforests.metadata import MetadataReader, Station, SurveyYear
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData

logger = logging.getLogger(__name__)


class ReadICPForest(ReadUngriddedBase):
    #: version log of this class (for caching)
    __version__ = "0.5_" + ReadUngriddedBase.__baseversion__

    #: Name of dataset (OBS_ID)
    DATA_ID = const.ICPFORESTS_NAME

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [const.ICPFORESTS_NAME]

    TS_TYPE = "undefined"

    _FILEMASK = "dp_dem.csv"

    #: Temporal resolution codes that (so far) can be understood by pyaerocom
    TS_TYPE_CODES = {
        "1mn": "minutely",
        "1h": "hourly",
        "1d": "daily",
        "1w": "weekly",
        "1mo": "monthly",
        "mn": "minutely",
        "h": "hourly",
        "d": "daily",
        "w": "weekly",
        "mo": "monthly",
    }

    VAR_POSITION = {
        "wetoxs": 20,
        "wetoxn": 19,
        "wetrdn": 17,
        "dryoxs": 20,
        "dryoxn": 19,
        "dryrdn": 17,
        "depoxs": 20,
        "depoxn": 19,
        "deprdn": 17,
        "depna": 16,
        "depoxsf": 20,
        "depoxnf": 19,
        "deprdnf": 17,
        "depnaf": 16,
        "wetso4": 20,
        "wetno3": 19,
        "wetnh4": 17,
        "wetno2": 41,
        "wetna": 16,
        "wetcl": 18,
    }

    UNITS = {
        "wetoxs": "mg S m-2 d-1",
        "wetoxn": "mg N m-2 d-1",
        "wetrdn": "mg N m-2 d-1",
        "dryoxs": "mg S m-2 d-1",
        "dryoxn": "mg N m-2 d-1",
        "dryrdn": "mg N m-2 d-1",
        "depoxs": "mg S m-2 d-1",
        "depoxn": "mg N m-2 d-1",
        "deprdn": "mg N m-2 d-1",
        "depoxsf": "mg S m-2 d-1",
        "depoxnf": "mg N m-2 d-1",
        "deprdnf": "mg N m-2 d-1",
        "wetso4": "mg S m-2 d-1",
        "wetno3": "mg N m-2 d-1",
        "wetnh4": "mg N m-2 d-1",
        "wetno2": "mg N m-2 d-1",
        "wetna": "mg m-2 d-1",
        "depna": "mg m-2 d-1",
        "depnaf": "mg m-2 d-1",
        # unverified
        "wetcl": "mg m-2 d-1",
    }

    SEASALT_CORRECTION = {
        "wetoxs": 0.3338,
        "wetso4": 0.3338,
        "depoxs": 0.3338,
        "depoxsf": 0.3338,
    }
    SEASALT_FACTORS = {
        "wetna": 0.120,
        "wetcl": 0.103,
    }

    DEP_TYPES_TO_USE = ["Throughfall", "Bulk", "Wet-only"]

    QUALITY_LIMIT = 0.5

    MIN_YEAR = 1984
    MAX_YEAR = 2019

    def __init__(self, data_id=None, data_dir=None):
        super().__init__(data_id, data_dir)

        self.metadata = None

        self._file_dir = None
        self.files = None

        if data_dir is not None:
            self.metadata = MetadataReader(data_dir)

    @property
    def file_dir(self):
        """Directory containing EBAS NASA Ames files"""
        if self._file_dir is not None:
            return self._file_dir
        return os.path.join(self.data_dir)

    @file_dir.setter
    def file_dir(self, val):
        if not isinstance(val, str) or not os.path.exists(val):
            raise FileNotFoundError("Input directory does not exist")

        self._file_dir = val

    def read(self, vars_to_retrieve=None, files=[], first_file=None, last_file=None):
        """Method that reads list of files as instance of :class:`UngriddedData`

        Parameters
        ----------
        vars_to_retrieve : :obj:`list` or similar, optional,
            list containing variable IDs that are supposed to be read. If None,
            all variables in :attr:`PROVIDES_VARIABLES` are loaded
        files : :obj:`list`, optional
            list of files to be read. If None, then the file list is used that
            is returned on :func:`get_file_list`.
        first_file : :obj:`int`, optional
            index of first file in file list to read. If None, the very first
            file in the list is used
        last_file : :obj:`int`, optional
            index of last file in list to read. If None, the very last file
            in the list is used

        Returns
        -------
        UngriddedData
            instance of ungridded data object containing data from all files.
        """
        self.files = Path(self.data_dir).joinpath(self._FILEMASK)
        data = self.read_file(
            str(self.files),
            vars_to_retrieve,
        )

        return data

    @property
    def PROVIDES_VARIABLES(self):
        """List of variables that are provided by this dataset

        Note
        ----
        May be implemented as global constant in header
        """
        return list(self.VAR_POSITION.keys())

    @property
    def DEFAULT_VARS(self):
        """List containing default variables to read"""
        return list(self.VAR_POSITION.keys())

    def read_file(self, filename, vars_to_retrieve=None, last_line=None):
        """Read single file

        Parameters
        ----------
        filename : str
            string specifying filename
        vars_to_retrieve : :obj:`list` or similar, optional,
            list containing variable IDs that are supposed to be read. If None,
            all variables in :attr:`PROVIDES_VARIABLES` are loaded
        last_line : int
            last line number to read
            Used to speed up testing

        Returns
        -------
        :obj:`dict` or :obj:`StationData`, or other...
            imported data in a suitable format that can be handled by
            :func:`read` which is supposed to append the loaded results from
            this method (which reads one datafile) to an instance of
            :class:`UngriddedData` for all files.
        """
        stations: dict[str, dict[str, Station]] = {}
        if self.metadata is None:
            if self.data_dir is None:
                raise ValueError(f"Data Dir is not read yet")

            self.metadata = MetadataReader(self.data_dir)

        if vars_to_retrieve is None:
            vars_to_retrieve = self.PROVIDES_VARIABLES

        with open(filename, "r") as f:
            f.readline()

            for line_nr, line in tqdm(enumerate(f)):
                words = line.split(";")
                year = int(words[0])
                country_code = int(words[1])
                partner_code = int(words[2])
                plot_code = int(words[3])
                sampler_code = int(words[9])
                q_flag = int(words[44]) if words[44] != "" else 0

                # 8 is the code for "do not use"
                if (
                    self.metadata.deposition_type[sampler_code] not in self.DEP_TYPES_TO_USE
                    or sampler_code == 8
                ):
                    continue

                sampler_type = self.metadata.deposition_type[sampler_code]

                period = int(words[6])
                start = words[4]
                stop = words[5]

                quantity = words[47]
                if quantity == "":  # or quantity == "0":
                    continue
                else:
                    quantity = float(quantity)

                try:
                    self.metadata.plots.plots[country_code]
                    self.metadata.plots.plots[country_code][plot_code]
                    self.metadata.plots.plots[country_code][plot_code][sampler_code]
                except KeyError:
                    logger.warning(
                        f"Some metadata is missing for {country_code=}, {plot_code=}, {sampler_code=}. Skipping"
                    )
                    continue

                try:
                    self.metadata.plots.plots[country_code][plot_code][sampler_code].survey_years[
                        year
                    ]
                except KeyError as e:
                    logger.warning(
                        f"Year {year} can't be found for {country_code=}, {plot_code=}, {sampler_code=}. Only years found are {self.metadata.plots.plots[country_code][plot_code][sampler_code].survey_years.keys()}"
                    )
                    continue
                days, dtime, ts_type = self._get_days_date_ts_type(
                    year, country_code, plot_code, sampler_code, period, start, stop
                )

                if days is None or dtime is None or ts_type is None:
                    continue
                station_name = Station.get_station_name(country_code, plot_code, sampler_code)

                if station_name not in stations:
                    stations[station_name] = {}
                if ts_type not in stations[station_name]:
                    lat, lon, alt = self.metadata.plots.get_position(
                        year, country_code, plot_code, sampler_code
                    )
                    stations[station_name][ts_type] = Station(
                        country_code, plot_code, sampler_code, lat, lon, alt, partner_code, ts_type
                    )

                for species in vars_to_retrieve:
                    conc = self._get_species_conc(words[self.VAR_POSITION[species]], species)

                    # Sea-salt correction
                    # The factor self.SEASALT_CORRECTION[species] is the factor use to go from mg/L to mg S/L (for sulpher)
                    if species in self.SEASALT_CORRECTION:
                        na_factor = (
                            self._get_species_conc(words[self.VAR_POSITION["wetna"]], "wetna")
                            * self.SEASALT_FACTORS["wetna"]
                            * self.SEASALT_CORRECTION[species]
                        )
                        cl_factor = (
                            self._get_species_conc(words[self.VAR_POSITION["wetcl"]], "wetcl")
                            * self.SEASALT_FACTORS["wetcl"]
                            * self.SEASALT_CORRECTION[species]
                        )
                        seasalt_correction = min(na_factor, cl_factor)

                        if seasalt_correction > conc and conc > 0:
                            logger.warning(
                                f"Seasalt correction {seasalt_correction} is larger than concentration {conc} for {species}"
                            )
                        if conc > 0:
                            conc -= seasalt_correction

                    # Unit correction
                    conc *= quantity / days

                    stations[station_name][ts_type].add_measurement(
                        species, dtime, conc, self.UNITS[species], q_flag
                    )

        station_datas = []
        for station_name in stations:
            for ts_type in stations[station_name]:
                station = stations[station_name][ts_type]
                station_data = StationData()
                station_data.var_info = BrowseDict(**station.var_info)
                for species in station.data.keys():
                    station_data[species] = self._clean_data_with_flags(
                        station.data[species],
                        station.dtime[species],
                        station.flags[species],
                        species,
                    )
                    station_data.dtime = station.dtime[species]

                station_data.country = station.country

                # Needs to convert coordinates to correct type!
                station_data.station_coords = {
                    "latitude": station.lat,
                    "longitude": station.lon,
                    "altitude": station.alt,
                }

                station_data.latitude = station.lat
                station_data.longitude = station.lon
                station_data.altitude = station.alt

                station_data.filename = filename
                station_data.ts_type = station.ts_type
                station_data.ts_type_src = station.ts_type
                station_data.station_name = station.station_name
                station_data._append_meta_item("sampler_type", station.sampler_type)

                station_data.data_id = self.data_id

                station_datas.append(station_data)
        return UngriddedData.from_station_data(station_datas, add_meta_keys="sampler_type")

    def _get_species_conc(self, conc_str: str, species: str) -> float:
        conc = float(conc_str) if conc_str != "" else np.nan
        if conc == -1:
            logger.warning(f"Value for {species} found to be {conc}")
            conc = np.nan
        return conc

    def _get_days_date_ts_type(
        self,
        year: int,
        country_code: int,
        plot_code: int,
        sampler_code: int,
        period: int,
        start: str | datetime,
        stop: str | datetime,
    ) -> Tuple[float | None, datetime | None, str | None]:
        if start != "" and stop != "":
            if isinstance(start, str):
                start = datetime.strptime(start, "%Y-%m-%d")
            if isinstance(stop, str):
                stop = datetime.strptime(stop, "%Y-%m-%d")

            if (stop - start).days <= 0:
                return None, None, None

            return (stop - start).days, start, self._get_tstype(start, stop)

        if self.metadata is None:
            raise ValueError(f"Metadata is not read yet")

        try:
            days = self.metadata.plots.get_days(year, country_code, plot_code, sampler_code)
        except ValueError as e:
            logger.warning(repr(e))
            return None, None, None

        if days == 0:
            return None, None, None

        try:
            dtime = self.metadata.plots.get_date(
                year, country_code, plot_code, sampler_code, period
            )
        except ValueError:
            return None, None, None

        ts_type = self.metadata.plots.get_ts_type(year, country_code, plot_code, sampler_code)

        return days, dtime, ts_type

    def _get_tstype(self, start: datetime, stop: datetime) -> str:
        days = (stop - start).days

        return SurveyYear.get_tstype(days)

    def _clean_data_with_flags(
        self,
        data: list[float],
        time: list[datetime],
        flags: list[int],
        species: str,
    ) -> list[float]:
        data_array = np.array(data)
        flags_array = np.array(flags)
        years = np.array([i.year for i in time])

        for year in range(self.MIN_YEAR, self.MAX_YEAR):
            yr_flags = flags_array[np.where(years == year)]
            # yr_flags can have 0 length
            try:
                quality = np.sum(np.where(yr_flags == 0)) / len(yr_flags)
                if quality < self.QUALITY_LIMIT:
                    logger.warning(
                        f"Quailty of {quality} found for {species} in year {year}. Setting data this year to NaN"
                    )
                    data_array[np.where(years == year)] = np.nan

            except RuntimeWarning:
                # yr_flags has 0 length
                logger.warning(f"No data for species {species} in year {year}.")

        return list(data_array)



from __future__ import annotations

import getpass
from pathlib import Path

import pytest

import pyaerocom.config as testmod
from pyaerocom import const
from pyaerocom.config import ALL_REGION_NAME, Config
from pyaerocom.data import resources
from pyaerocom.grid_io import GridIO
from pyaerocom.varcollection import VarCollection
from tests.conftest import lustre_avail

USER = getpass.getuser()


@pytest.fixture()
def config_file(file: str | None, tmp_path: Path) -> str | None:
    if file is None:
        return None

    if file.casefold() == "default":
        assert resources.is_resource("pyaerocom.data", "paths.ini")
        with resources.path("pyaerocom.data", "paths.ini") as path:
            return str(path)

    if file.casefold() == "wrong_suffix":
        path = tmp_path / "paths.txt"
        path.write_text("")
        assert path.exists()
        return str(path)

    return file


@pytest.fixture()
def local_db(tmp_path: Path) -> Path:
    """temporary path to DB file structure"""
    path = tmp_path / "data"
    (path / "modeldata").mkdir(parents=True)
    (path / "obsdata").mkdir()
    assert path.is_dir()
    return path


@pytest.fixture(scope="module")
def empty_cfg():
    cfg = testmod.Config(try_infer_environment=False)
    return cfg


def test_Config_ALL_DATABASE_IDS(empty_cfg):
    assert empty_cfg.ALL_DATABASE_IDS == ["metno", "users-db", "local-db"]


@pytest.mark.parametrize(
    "file,try_infer_environment",
    [
        (None, False),
        (None, True),
        ("default", False),
    ],
)
def test_Config___init__(config_file: str, try_infer_environment: bool):
    testmod.Config(config_file, try_infer_environment)


@pytest.mark.parametrize(
    "file,exception,error",
    [
        pytest.param(
            "wrong_suffix",
            ValueError,
            "Need path to an ini file for input config_file",
            id="wrong file extension",
        ),
        pytest.param(
            f"/home/{USER}/blaaaa.ini",
            FileNotFoundError,
            f"input config file does not exist /home/{USER}/blaaaa.ini",
            id="no such file",
        ),
    ],
)
def test_Config___init___error(config_file: str, exception: type[Exception], error: str):
    with pytest.raises(exception) as e:
        testmod.Config(config_file, False)
    assert str(e.value) == error


def test_Config__infer_config_from_basedir(local_db: Path):
    cfg = testmod.Config(try_infer_environment=False)
    res = cfg._infer_config_from_basedir(local_db)
    assert res[1] == "local-db"


def test_Config__infer_config_from_basedir_error():
    cfg = testmod.Config(try_infer_environment=False)
    with pytest.raises(FileNotFoundError):
        cfg._infer_config_from_basedir("/blaaa")


def test_Config_has_access_lustre():
    cfg = testmod.Config(try_infer_environment=False)
    assert not cfg.has_access_lustre


def test_Config_has_access_users_database():
    cfg = testmod.Config(try_infer_environment=False)
    assert not cfg.has_access_users_database


@lustre_avail
@pytest.mark.parametrize(
    "cfg_id,basedir,init_obslocs_ungridded,init_data_search_dirs,data_searchdirno",
    [
        ("metno", None, False, False, 0),
        ("metno", None, True, False, 0),
        ("metno", None, True, True, 0),
        ("metno", f"/home/{USER}", True, True, 2),
        ("users-db", None, False, False, 0),
    ],
)
def test_Config_read_config(
    cfg_id, basedir, init_obslocs_ungridded, init_data_search_dirs, data_searchdirno
):
    cfg = testmod.Config(try_infer_environment=False)
    cfg_file = cfg._config_files[cfg_id]
    assert Path(cfg_file).exists()
    cfg.read_config(cfg_file, basedir, init_obslocs_ungridded, init_data_search_dirs)
    assert len(cfg.DATA_SEARCH_DIRS) == data_searchdirno
    assert len(cfg.OBSLOCS_UNGRIDDED) == 0
    assert Path(cfg.OUTPUTDIR).exists()
    assert Path(cfg.COLOCATEDDATADIR).exists()
    assert Path(cfg.CACHEDIR).exists()


def test_empty_class_header(empty_cfg):
    cfg = empty_cfg
    assert cfg.AERONET_SUN_V2L15_AOD_DAILY_NAME == "AeronetSunV2Lev1.5.daily"
    assert cfg.AERONET_SUN_V2L15_AOD_ALL_POINTS_NAME == "AeronetSun_2.0_NRT"
    assert cfg.AERONET_SUN_V2L2_AOD_DAILY_NAME == "AeronetSunV2Lev2.daily"
    assert cfg.AERONET_SUN_V2L2_AOD_ALL_POINTS_NAME == "AeronetSunV2Lev2.AP"
    assert cfg.AERONET_SUN_V2L2_SDA_DAILY_NAME == "AeronetSDAV2Lev2.daily"
    assert cfg.AERONET_SUN_V2L2_SDA_ALL_POINTS_NAME == "AeronetSDAV2Lev2.AP"
    assert cfg.AERONET_INV_V2L15_DAILY_NAME == "AeronetInvV2Lev1.5.daily"
    assert cfg.AERONET_INV_V2L15_ALL_POINTS_NAME == "AeronetInvV2Lev1.5.AP"
    assert cfg.AERONET_INV_V2L2_DAILY_NAME == "AeronetInvV2Lev2.daily"
    assert cfg.AERONET_INV_V2L2_ALL_POINTS_NAME == "AeronetInvV2Lev2.AP"
    assert cfg.AERONET_SUN_V3L15_AOD_DAILY_NAME == "AeronetSunV3Lev1.5.daily"
    assert cfg.AERONET_SUN_V3L15_AOD_ALL_POINTS_NAME == "AeronetSunV3Lev1.5.AP"
    assert cfg.AERONET_SUN_V3L2_AOD_DAILY_NAME == "AeronetSunV3Lev2.daily"
    assert cfg.AERONET_SUN_V3L2_AOD_ALL_POINTS_NAME == "AeronetSunV3Lev2.AP"
    assert cfg.AERONET_SUN_V3L15_SDA_DAILY_NAME == "AeronetSDAV3Lev1.5.daily"
    assert cfg.AERONET_SUN_V3L15_SDA_ALL_POINTS_NAME == "AeronetSDAV3Lev1.5.AP"
    assert cfg.AERONET_SUN_V3L2_SDA_DAILY_NAME == "AeronetSDAV3Lev2.daily"
    assert cfg.AERONET_SUN_V3L2_SDA_ALL_POINTS_NAME == "AeronetSDAV3Lev2.AP"
    assert cfg.AERONET_INV_V3L15_DAILY_NAME == "AeronetInvV3Lev1.5.daily"
    assert cfg.AERONET_INV_V3L2_DAILY_NAME == "AeronetInvV3Lev2.daily"
    assert cfg.EBAS_MULTICOLUMN_NAME == "EBASMC"
    assert cfg.EEA_NAME == "EEAAQeRep"
    assert cfg.EARLINET_NAME == "EARLINET"
    assert cfg.GAWTADSUBSETAASETAL_NAME == "GAWTADsubsetAasEtAl"
    assert cfg.DMS_AMS_CVO_NAME == "DMS_AMS_CVO"
    assert cfg.MEP_NAME == "MEP"
    assert cfg.EBAS_DB_LOCAL_CACHE
    assert cfg.MIN_YEAR == 0
    assert cfg.MAX_YEAR == 20000
    assert cfg.STANDARD_COORD_NAMES == ["latitude", "longitude", "altitude"]
    assert cfg.DEFAULT_VERT_GRID_DEF == dict(lower=0, upper=15000, step=250)
    assert cfg.RH_MAX_PERCENT_DRY == 40
    assert cfg.DEFAULT_REG_FILTER == f"{ALL_REGION_NAME}-wMOUNTAINS"
    assert cfg.OBS_MIN_NUM_RESAMPLE == dict(
        yearly=dict(monthly=3),
        monthly=dict(daily=7),
        daily=dict(hourly=6),
        hourly=dict(minutely=15),
    )
    from pyaerocom import obs_io

    assert cfg.OBS_ALLOW_ALT_WAVELENGTHS == obs_io.OBS_ALLOW_ALT_WAVELENGTHS

    assert cfg.OBS_WAVELENGTH_TOL_NM == obs_io.OBS_WAVELENGTH_TOL_NM

    assert cfg.CLIM_START == 2005
    assert cfg.CLIM_STOP == 2015
    assert cfg.CLIM_FREQ == "daily"
    assert cfg.CLIM_RESAMPLE_HOW == "mean"
    assert cfg.CLIM_MIN_COUNT == dict(daily=30, monthly=5)

    # names for the satellite data sets
    assert cfg.SENTINEL5P_NAME == "Sentinel5P"
    assert cfg.AEOLUS_NAME == "AeolusL2A"

    assert cfg.OLD_AEROCOM_REGIONS == [
        ALL_REGION_NAME,
        "ASIA",
        "AUSTRALIA",
        "CHINA",
        "EUROPE",
        "INDIA",
        "NAFRICA",
        "SAFRICA",
        "SAMERICA",
        "NAMERICA",
    ]

    assert cfg.URL_HTAP_MASKS == "https://pyaerocom.met.no/pyaerocom-suppl/htap_masks/"

    assert cfg.HTAP_REGIONS == [
        "PAN",
        "EAS",
        "NAF",
        "MDE",
        "LAND",
        "SAS",
        "SPO",
        "OCN",
        "SEA",
        "RBU",
        "EEUROPE",
        "NAM",
        "WEUROPE",
        "SAF",
        "USA",
        "SAM",
        "EUR",
        "NPO",
        "MCA",
    ]

    assert cfg.RM_CACHE_OUTDATED

    #: Name of the file containing the revision string of an obs data network
    assert cfg.REVISION_FILE == "Revision.txt"

    #: timeout to check if one of the supported server locations can be
    #: accessed
    assert cfg.SERVER_CHECK_TIMEOUT == 1  # s

    assert cfg._outhomename == "MyPyaerocom"

    with resources.path("pyaerocom.data", "paths.ini") as path:
        assert cfg._config_files["metno"] == cfg._config_ini_lustre == str(path)

    with resources.path("pyaerocom.data", "paths_user_server.ini") as path:
        assert cfg._config_files["users-db"] == cfg._config_ini_user_server == str(path)

    with resources.path("pyaerocom.data", "paths_local_database.ini") as path:
        assert cfg._config_files["local-db"] == cfg._config_ini_localdb == str(path)

    assert cfg._check_subdirs_cfg == {
        "metno": "aerocom",
        "users-db": "AMAP",
        "local-db": "modeldata",
    }

    with resources.path("pyaerocom.data", "variables.ini") as path:
        assert cfg._var_info_file == str(path)
    with resources.path("pyaerocom.data", "coords.ini") as path:
        assert cfg._coords_info_file == str(path)

    dbdirs = {
        "lustre/storeB/project": "metno",
        "metno/aerocom_users_database": "users-db",
        "MyPyaerocom/data": "local-db",
    }
    for sd, name in dbdirs.items():
        assert sd in cfg._DB_SEARCH_SUBDIRS
        assert cfg._DB_SEARCH_SUBDIRS[sd] == name

    assert cfg.DONOTCACHEFILE == None

    assert cfg.ERA5_SURFTEMP_FILENAME == "era5.msl.t2m.201001-201012.nc"

    assert cfg._LUSTRE_CHECK_PATH == "/project/aerocom/aerocom1/"


def test_empty_init(empty_cfg):
    cfg = empty_cfg
    assert cfg._cache_basedir is None
    assert cfg._outputdir is None
    assert cfg._colocateddatadir is None
    assert cfg._filtermaskdir is None
    assert cfg._local_tmp_dir is None
    assert cfg._downloaddatadir is None
    assert cfg._confirmed_access == []
    assert cfg._rejected_access == []

    # Options
    assert cfg._caching_active is True

    assert cfg._var_param is None
    assert cfg._coords is None

    # Attributes that are used to store search directories
    assert cfg.OBSLOCS_UNGRIDDED == {}
    assert cfg.OBS_UNGRIDDED_POST == {}
    assert cfg.SUPPLDIRS == {}
    assert cfg._search_dirs == []

    assert cfg.WRITE_FILEIO_ERR_LOG is True

    assert cfg.last_config_file is None
    assert cfg._ebas_flag_info is None
    from pyaerocom.grid_io import GridIO

    #: Settings for reading and writing of gridded data
    assert isinstance(cfg.GRID_IO, GridIO)


def test_default_config_HOMEDIR():
    assert Path(const.HOMEDIR) == Path("~").expanduser()
    assert const.HOMEDIR.endswith("/")


@lustre_avail
def test_default_config():
    cfg = Config()

    mypydir = Path(cfg.HOMEDIR).resolve() / "MyPyaerocom"
    assert Path(cfg.OUTPUTDIR) == Path(cfg._outputdir) == mypydir
    assert Path(cfg.CACHEDIR) == mypydir / f"_cache/{USER}"
    assert Path(cfg.COLOCATEDDATADIR) == Path(cfg._colocateddatadir) == mypydir / "colocated_data"
    assert Path(cfg.FILTERMASKKDIR) == Path(cfg._filtermaskdir) == mypydir / "filtermasks"
    assert Path(cfg.LOCAL_TMP_DIR) == Path(cfg._local_tmp_dir) == mypydir / "tmp"
    assert Path(cfg.DOWNLOAD_DATADIR) == Path(cfg._downloaddatadir) == mypydir / "data"

    assert cfg._caching_active
    assert cfg.CACHING

    assert isinstance(cfg.VARS, VarCollection)
    assert cfg.VARS is cfg.VAR_PARAM
    assert isinstance(cfg._var_param, VarCollection)

    assert isinstance(cfg.COORDINFO, VarCollection)
    assert cfg._coords is cfg.COORDINFO

    assert cfg.DATA_SEARCH_DIRS is cfg._search_dirs

    assert cfg.WRITE_FILEIO_ERR_LOG

    assert isinstance(cfg.GRID_IO, GridIO)



from __future__ import annotations

from pathlib import Path

import pytest

from pyaerocom import const
from pyaerocom.plugins.mep.reader import ReadMEP
from tests.conftest import lustre_unavail

try:
    MEP_PATH = Path(const.OBSLOCS_UNGRIDDED[const.MEP_NAME])
except KeyError:
    pytest.mark.skip(reason=f"MEP path not initialised due to non existence in CI")

STATION_NAMES = ("1478A", "2706A", "3377A")

VARS_DEFAULT = {"concco", "concno2", "conco3", "concpm10", "concpm25", "concso2"}
VARS_PROVIDED = VARS_DEFAULT | {"vmro3", "vmro3max", "vmrno2"}


@lustre_unavail
@pytest.fixture(scope="module")
def reader() -> ReadMEP:
    return ReadMEP(data_dir=str(MEP_PATH))


@lustre_unavail
@pytest.fixture()
def station_files(station: str) -> list[Path]:
    files = sorted(MEP_PATH.rglob(f"mep-rd-{station}-*.nc"))
    assert files, f"no files for {station}"
    return files


@lustre_unavail
def test_DATASET_NAME(reader: ReadMEP):
    assert reader.DATASET_NAME == "MEP"


@lustre_unavail
def test_DEFAULT_VARS(reader: ReadMEP):
    assert set(reader.DEFAULT_VARS) >= VARS_DEFAULT


@lustre_unavail
def test_files(reader: ReadMEP):
    assert reader.files, "no stations files found"
    assert len(reader.files) >= 3, "found less files than expected"


@lustre_unavail
def test_FOUND_FILES(reader: ReadMEP):
    assert reader.FOUND_FILES, "no stations files found"
    assert len(reader.FOUND_FILES) >= 3, "found less files than expected"


@lustre_unavail
@pytest.mark.parametrize("station", STATION_NAMES)
def test_stations(reader: ReadMEP, station: str):
    assert reader.stations()[station], f"no {station} station files"


@lustre_unavail
def test_PROVIDES_VARIABLES(reader: ReadMEP):
    assert set(reader.PROVIDES_VARIABLES) >= VARS_PROVIDED


@lustre_unavail
@pytest.mark.parametrize("station", STATION_NAMES)
def test_read_file(reader: ReadMEP, station_files: list[str]):
    data = reader.read_file(station_files[-1])
    assert set(data.contains_vars) == VARS_DEFAULT


@lustre_unavail
def test_read_file_error(reader: ReadMEP):
    bad_station_file = "not-a-file.nc"
    with pytest.raises(ValueError) as e:
        reader.read_file(bad_station_file)
    assert str(e.value) == f"missing {bad_station_file}"


@lustre_unavail
@pytest.mark.parametrize("station", STATION_NAMES)
def test_read(reader: ReadMEP, station_files: list[str]):
    data = reader.read(VARS_PROVIDED, station_files, first_file=0, last_file=5)
    assert set(data.contains_vars) == VARS_PROVIDED


@lustre_unavail
def test_read_error(reader: ReadMEP):
    bad_variable_name = "not-a-variable"
    with pytest.raises(ValueError) as e:
        reader.read((bad_variable_name,))
    assert str(e.value) == f"Unsupported variables: {bad_variable_name}"


@lustre_unavail
def test_reader_gives_correct_mep_path(reader: ReadMEP):
    assert str(MEP_PATH) == reader.data_dir



from pyaerocom.aeroval.modelmaps_helpers import _jsdate_list, griddeddata_to_jsondict


def test__jsdate_list(data_tm5):
    vals = _jsdate_list(data_tm5)
    assert len(vals) == 12
    assert vals[0] == 1263513600000
    assert vals[-1] == 1292371200000


def test_griddeddata_to_jsondict(data_tm5):
    result = griddeddata_to_jsondict(data_tm5)
    assert isinstance(result, dict)
    assert "metadata" in result
    assert len(result["metadata"]) == 2
    assert "data" in result
    data = result["data"]
    assert "time" in data
    assert len(data["time"]) == 12
    assert len(data) == 2593
    del data["time"]
    pixel = list(data.values())[0]
    assert "lat" in pixel
    assert isinstance(pixel["lat"], float)
    assert "lon" in pixel
    assert isinstance(pixel["lon"], float)
    assert "data" in pixel
    assert len(pixel["data"]) == 12



from pathlib import Path

import numpy as np
import pytest

from pyaerocom.io.read_aeronet_sdav2 import ReadAeronetSdaV2
from tests.conftest import TEST_RTOL, lustre_unavail


@lustre_unavail
def test_load_berlin_AeroSdaV2L2D():
    reader = ReadAeronetSdaV2()
    files = reader.find_in_file_list("*Berlin*")
    assert len(files) == 1
    assert Path(files[0]).name == "920801_180519_Berlin_FUB.ONEILL_20"

    test_vars = ["od870aer", "ang4487aer", "od550aer", "od550gt1aer", "od550lt1aer"]

    data = reader.read_file(files[0], vars_to_retrieve=test_vars)

    assert all(x in data for x in test_vars)

    # more than 100 timestamps
    assert all(len(data[x]) > 100 for x in test_vars)

    assert isinstance(data["dtime"][0], np.datetime64)
    assert data["dtime"][0] == np.datetime64("2014-07-06T00:00:00")

    means = [data[var].mean() for var in test_vars]
    desired = [
        0.0671392659574468,
        1.5027900015754372,
        0.1338938495016503,
        0.02333982521802314,
        0.11055405137562464,
    ]
    assert means == pytest.approx(desired, rel=TEST_RTOL)



import pytest

from pyaerocom.aeroval.setupclasses import EvalSetup
from pyaerocom.exceptions import EvalEntryNameError
from tests.fixtures.aeroval.cfg_test_exp1 import CFG


@pytest.mark.parametrize(
    "cfg,update",
    [
        ("cfgexp1", dict()),
        ("cfgexp1", dict(obs_cfg=dict(OBS=CFG["obs_cfg"]["AERONET-Sun"]))),
    ],
)
def test_EvalSetup___init__(eval_config: dict, update: dict):
    eval_config.update(update)
    EvalSetup(**eval_config)


@pytest.mark.parametrize(
    "cfg,update,error",
    [
        (
            "cfgexp1",
            dict(model_cfg=dict(WRONG_MODEL=CFG["model_cfg"]["TM5-AP3-CTRL"])),
            "Invalid name: WRONG_MODEL",
        ),
        (
            "cfgexp1",
            dict(obs_cfg=dict(WRONG_OBS=CFG["obs_cfg"]["AERONET-Sun"])),
            "Invalid name: WRONG_OBS",
        ),
        (
            "cfgexp1",
            dict(obs_cfg=dict(OBS=dict(web_interface_name="WRONG_OBS"))),
            "Invalid name: WRONG_OBS",
        ),
    ],
)
def test_EvalSetup___init__INVALID_ENTRY_NAMES(eval_config: dict, update: dict, error: str):
    eval_config.update(update)
    with pytest.raises(EvalEntryNameError) as e:
        EvalSetup(**eval_config)
    assert error in str(e.value)



"""
read Aeronet inversion V2 data
"""
import os
import re

import numpy as np
import pandas as pd

from pyaerocom import const
from pyaerocom.aux_var_helpers import calc_abs550aer, calc_od550aer
from pyaerocom.io.readaeronetbase import ReadAeronetBase
from pyaerocom.stationdata import StationData


class ReadAeronetInvV2(ReadAeronetBase):
    """Interface for reading Aeronet inversion V2 Level 1.5 and 2.0 data

    Note
    ----
    Maintenance for AERONET version 2 data is deprecated since pyaerocom
    version 0.11.0. Please use version 3 data and associated pyaerocom readers.
    """

    #: Mask for identifying datafiles
    _FILEMASK = "*.dubovikday"

    #: version log of this class (for caching)
    __version__ = "0.08_" + ReadAeronetBase.__baseversion__

    #: Name of dataset (OBS_ID)
    DATA_ID = const.AERONET_INV_V2L2_DAILY_NAME

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [const.AERONET_INV_V2L2_DAILY_NAME, const.AERONET_INV_V2L15_DAILY_NAME]

    #: dictionary assigning temporal resolution flags for supported datasets
    #: that are provided in a defined temporal resolution
    TS_TYPES = {
        const.AERONET_INV_V2L2_DAILY_NAME: "daily",
        const.AERONET_INV_V2L15_DAILY_NAME: "daily",
    }

    #: default variables for read method
    DEFAULT_VARS = ["ssa675aer", "ssa440aer", "ssa870aer", "ssa1020aer", "abs550aer", "od550aer"]

    #: value corresponding to invalid measurement
    NAN_VAL = -9999.0

    #: dictionary specifying the file column names (values) for each Aerocom
    #: variable (keys)
    VAR_NAMES_FILE = {}
    VAR_NAMES_FILE["ssa440aer"] = "SSA440-T"
    VAR_NAMES_FILE["ssa675aer"] = "SSA675-T"
    VAR_NAMES_FILE["ssa870aer"] = "SSA870-T"
    VAR_NAMES_FILE["ssa1020aer"] = "SSA1020-T"
    VAR_NAMES_FILE["od440aer"] = "AOTExt440-T"  # derived from inversion
    VAR_NAMES_FILE["ang4487aer"] = "870-440AngstromParam.[AOTExt]-Total"
    VAR_NAMES_FILE["abs440aer"] = "AOTAbsp440-T"
    VAR_NAMES_FILE["angabs4487aer"] = "870-440AngstromParam.[AOTAbsp]"

    #: OPTIONAL: dictionary specifying alternative column names for variables
    #: defined in :attr:`VAR_NAMES_FILE`. Check attribute _alt_vars_cols after
    #: running read().
    ALT_VAR_NAMES_FILE = {}
    ALT_VAR_NAMES_FILE["ssa440aer"] = ["SSA439-T", "SSA441-T", "SSA438-T", "SSA437-T", "SSA442-T"]
    ALT_VAR_NAMES_FILE["ssa675aer"] = [
        "SSA676-T",
        "SSA673-T",
        "SSA674-T",
        "SSA669-T",
        "SSA677-T",
        "SSA668-T",
        "SSA672-T",
    ]
    ALT_VAR_NAMES_FILE["ssa870aer"] = [
        "SSA871-T",
        "SSA869-T",
        "SSA868-T",
        "SSA873-T",
        "SSA867-T",
        "SSA872-T",
    ]
    ALT_VAR_NAMES_FILE["ssa1020aer"] = ["SSA1022-T", "SSA1016-T", "SSA1018-T"]
    ALT_VAR_NAMES_FILE["od440aer"] = [
        "AOTExt439-T",
        "AOTExt441-T",
        "AOTExt438-T",
        "AOTExt437-T",
        "AOTExt442-T",
    ]
    ALT_VAR_NAMES_FILE["abs440aer"] = [
        "AOTAbsp439-T",
        "AOTAbsp441-T",
        "AOTAbsp438-T",
        "AOTAbsp437-T",
        "AOTAbsp442-T",
    ]

    #: dictionary containing information about additionally required variables
    #: for each auxiliary variable (i.e. each variable that is not provided
    #: by the original data but computed on import)
    AUX_REQUIRES = {
        "abs550aer": ["abs440aer", "angabs4487aer"],
        "od550aer": ["od440aer", "ang4487aer"],
    }

    #: Functions that are used to compute additional variables (i.e. one
    #: for each variable defined in AUX_REQUIRES)
    AUX_FUNS = {"abs550aer": calc_abs550aer, "od550aer": calc_od550aer}

    #: dictionary specifying the file column names (values) for each
    #: metadata key (cf. attributes of :class:`StationData`, e.g.
    #: 'station_name', 'longitude', 'latitude', 'altitude')
    META_NAMES_FILE = {}
    META_NAMES_FILE["data_quality_level"] = "DATA_TYPE"
    META_NAMES_FILE["date"] = "Date(dd-mm-yyyy)"
    META_NAMES_FILE["time"] = "Time(hh:mm:ss)"
    META_NAMES_FILE["day_of_year"] = "Julian_Day"

    #: List of variables that are provided by this dataset (will be extended
    #: by auxiliary variables on class init, for details see __init__ method of
    #: base class ReadUngriddedBase)
    PROVIDES_VARIABLES = list(VAR_NAMES_FILE)

    def read_file(self, filename, vars_to_retrieve=None, vars_as_series=False):
        """Read Aeronet file containing results from v2 inversion algorithm

        Parameters
        ----------
        filename : str
            absolute path to filename to read
        vars_to_retrieve : list
            list of str with variable names to read
        vars_as_series : bool
            if True, the data columns of all variables in the result dictionary
            are converted into pandas Series objects

        Returns
        -------
        StationData
            dict-like object containing results

        Example
        -------
        >>> import pyaerocom.io as pio
        >>> obj = pio.read_aeronet_invv2.ReadAeronetInvV2()
        >>> files = obj.get_file_list()
        >>> filedata = obj.read_file(files[0])
        """
        # implemented in base class
        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(vars_to_retrieve)

        # create empty data object (is dictionary with extended functionality)
        data_out = StationData()

        # create empty arrays for meta information
        for item in self.META_NAMES_FILE:
            data_out[item] = []

        # create empty arrays for all variables that are supposed to be read
        # from file
        for var in vars_to_read:
            data_out[var] = []

        # Iterate over the lines of the file
        self.logger.info(f"Reading file {filename}")

        with open(filename) as in_file:
            # get rid of the first com,a seperated string element...
            c_dummy = ",".join(in_file.readline().strip().split(",")[1:])
            # re.split(r'=|\,',c_dummy)
            i_dummy = iter(re.split(r"=|\,", c_dummy.rstrip()))
            dict_loc = dict(zip(i_dummy, i_dummy))

            data_out["latitude"] = float(dict_loc["lat"])
            data_out["longitude"] = float(dict_loc["long"])
            data_out["altitude"] = float(dict_loc["elev"])
            data_out["station_name"] = dict_loc["Locations"]
            data_out["PI"] = dict_loc["PI"]
            data_out["PI_email"] = dict_loc["Email"]

            data_out["ts_type"] = self.TS_TYPE

            # skip next two lines
            self.logger.info(f"Skipping line:\n{in_file.readline()}")
            self.logger.info(f"Skipping line:\n{in_file.readline()}")

            col_index_str = in_file.readline()
            if col_index_str != self._last_col_index_str:
                self.logger.info("Header has changed, reloading col_index map")
                self._update_col_index(col_index_str)
            col_index = self.col_index

            # dependent on the station, some of the required input variables
            # may not be provided in the data file. These will be ignored
            # in the following list that iterates over all data rows and will
            # be filled below, with vectors containing NaNs after the file
            # reading loop
            vars_available = {}
            col_names = {}
            for var in vars_to_read:
                if var in col_index:
                    idx = col_index[var]
                    vars_available[var] = idx
                    col_names[var] = self._last_col_order[idx]
                else:
                    self.logger.warning(
                        f"Variable {var} not available in file {os.path.basename(filename)}"
                    )
            data_out["col_names"] = col_names
            for line in in_file:
                # process line
                dummy_arr = line.strip().split(self.COL_DELIM)

                # This uses the numpy datestring64 functions that i.e. also
                # support Months as a time step for timedelta
                # Build a proper ISO 8601 UTC date string
                day, month, year = dummy_arr[col_index["date"]].split(":")
                datestring = "-".join([year, month, day])
                datestring = "T".join([datestring, dummy_arr[col_index["time"]]])
                # NOTE JGLISS: parsing timezone offset was removed on 22/2/19
                # since it is deprecated in recent numpy versions, for details
                # see https://www.numpy.org/devdocs/reference/arrays.datetime.html#changes-with-numpy-1-11
                # datestring = '+'.join([datestring, '00:00'])
                data_out["dtime"].append(np.datetime64(datestring))

                for var in self.META_NAMES_FILE:
                    val = dummy_arr[col_index[var]]
                    try:
                        # e.g. lon, lat, altitude
                        val = float(val)
                    except Exception:
                        pass
                    data_out[var].append(val)

                # copy the data fields that are available (rest will be filled
                # below)
                for var, idx in vars_available.items():
                    val = np.float_(dummy_arr[idx])
                    if val == self.NAN_VAL:
                        val = np.nan
                    data_out[var].append(val)

        # convert all lists to numpy arrays
        data_out["dtime"] = np.asarray(data_out["dtime"])

        for item in self.META_NAMES_FILE:
            data_out[item] = np.asarray(data_out[item])

        for var in vars_to_read:
            if var in vars_available:
                array = np.asarray(data_out[var])
            else:
                array = np.zeros(len(data_out["dtime"])) * np.nan
            data_out[var] = array

        # compute additional variables (if applicable)
        data_out = self.compute_additional_vars(data_out, vars_to_compute)

        if vars_as_series:
            for var in vars_to_read + vars_to_compute:
                if var in vars_to_retrieve:
                    data_out[var] = pd.Series(data_out[var], index=data_out["dtime"])
                else:
                    del data_out[var]

        return data_out



import numpy as np
import pytest

from pyaerocom.mathutils import (
    _nanmean_and_std,
    calc_statistics,
    estimate_value_range,
    exponent,
    is_strictly_monotonic,
    make_binlist,
)


@pytest.mark.parametrize("vmin, vmax, num", [(0, 1, 10), (0.345, 0.346, 100), (-2, -10, 5)])
def test_make_binlist(vmin, vmax, num):
    bins = make_binlist(vmin, vmax, num)
    assert isinstance(bins, list)
    assert bins[0] == vmin
    assert bins[-1] == vmax
    assert len(bins) == num + 1


@pytest.mark.parametrize(
    "inputval,result", [([1, 2], True), ([1], True), ([1, 2, 2], False), ([3, 2], False)]
)
def test_is_strictly_monotonic(inputval, result):
    assert is_strictly_monotonic(inputval) == result


@pytest.mark.parametrize("inputval, desired", [(0.01, -2), (4, 0), (234, 2)])
def test_exponent(inputval, desired):
    assert exponent(inputval) == desired


@pytest.mark.parametrize(
    "data,expected",
    [
        ([1], (1, 0)),
        (np.asarray([1]), (1, 0)),
        ([1, np.nan], (1, 0)),
        ([np.nan, np.nan], (np.nan, np.nan)),
        (np.random.normal(loc=3, scale=0.01, size=100000), (3, 0.01)),
        ([1, np.nan, 0, 2], (1, 0.816497)),
    ],
)
def test__nanmean_and_std(data, expected):
    assert _nanmean_and_std(data) == pytest.approx(expected, abs=0.001, rel=1e-2, nan_ok=True)


perfect_stats_num1_mean1 = {
    "totnum": 1,
    "num_valid": 1,
    "refdata_mean": 1,
    "refdata_std": 0,
    "data_mean": 1,
    "data_std": 0,
    "weighted": False,
    "rms": 0,
    "nmb": 0,
    "mnmb": 0,
    "mb": 0,
    "mab": 0,
    "fge": 0,
    "R": np.nan,
    "R_kendall": np.nan,
    "R_spearman": np.nan,
}
perfect_stats_num2_mean1 = perfect_stats_num1_mean1.copy()
perfect_stats_num2_mean1.update(totnum=2)

num_fakedata = 1000
zero_signal = np.zeros(num_fakedata)
noise = np.random.normal(loc=0, scale=0.01, size=num_fakedata)

idx = np.linspace(0, 2 * np.pi, num_fakedata)
sin_signal = np.sin(idx)


@pytest.mark.parametrize(
    "data,ref_data,expected",
    [
        ([1], [1], perfect_stats_num1_mean1),
        ([1, np.nan], [1, np.nan], perfect_stats_num2_mean1),
        (
            zero_signal,
            zero_signal,
            {
                "totnum": 1000.0,
                "num_valid": 1000.0,
                "refdata_mean": 0.0,
                "refdata_std": 0,
                "data_mean": 0.0,
                "data_std": 0,
                "weighted": False,
                "rms": 0.0,
                "R": np.nan,
                "R_spearman": np.nan,
                "R_kendall": np.nan,
                "nmb": 0,
                "mb": 0,
                "mab": 0,
                "mnmb": np.nan,
                "fge": np.nan,
            },
        ),
        (
            zero_signal,
            noise,
            {
                "totnum": 1000.0,
                "num_valid": 1000.0,
                "refdata_mean": 0.0,
                "refdata_std": 0,
                "data_mean": 0.0,
                "data_std": 0,
                "weighted": False,
                "rms": 0.0,
                "R": np.nan,
                "R_spearman": np.nan,
                "R_kendall": np.nan,
                "nmb": -1,
                "mb": 0,
                "mab": 0,
                "mnmb": -2,
                "fge": 2,
            },
        ),
        (
            sin_signal,
            sin_signal,
            {
                "totnum": 1000.0,
                "num_valid": 1000.0,
                "refdata_mean": 0.0,
                "refdata_std": 0.71,
                "data_mean": 0.0,
                "data_std": 0.71,
                "weighted": False,
                "rms": 0.0,
                "R": 1.0,
                "R_spearman": 1.0,
                "R_kendall": 1.0,
                "nmb": 0,
                "mb": 0,
                "mab": 0,
                "mnmb": np.nan,
                "fge": np.nan,
            },
        ),
    ],
)
def test_calc_statistics(data, ref_data, expected):
    stats = calc_statistics(data, ref_data)
    assert isinstance(stats, dict)
    assert len(stats) == len(expected)
    for key, val in expected.items():
        assert key in stats
        assert stats[key] == pytest.approx(val, abs=0.02, rel=0.01, nan_ok=True)


def test_calc_statistics_error():
    with pytest.raises(IndexError) as e:
        calc_statistics([1], [1, np.nan])
    assert str(e.value).startswith("boolean index did not match indexed array")


@pytest.mark.parametrize(
    "vmin,vmax,extend_percent,result",
    [
        (0, 1, 0, (0, 1)),
        (-0.012, 3.12345666, 0, (-0.02, 3.13)),
        (-0.012, 3.12345666, 5, (-0.2, 3.3)),
    ],
)
def test_estimate_value_range(vmin, vmax, extend_percent, result):
    vals = estimate_value_range(vmin, vmax, extend_percent)
    assert vals == pytest.approx(result, rel=1e-3)


def test_estimate_value_range_error():
    with pytest.raises(ValueError) as e:
        estimate_value_range(0, 0)
    assert str(e.value) == "vmax needs to exceed vmin"



from __future__ import annotations

import pytest

from pyaerocom.aeroval.helpers import (
    _check_statistics_periods,
    _get_min_max_year_periods,
    _period_str_to_timeslice,
    check_var_ranges_avail,
    make_dummy_model,
)
from pyaerocom.aeroval.setupclasses import EvalSetup
from pyaerocom.exceptions import VariableDefinitionError
from pyaerocom.griddeddata import GriddedData


@pytest.mark.parametrize(
    "dvar,var",
    [
        ("od550aer", "od550aer"),
        ("od550aer", "od550gt1aer"),
        ("pr", "prmm"),
        ("prmm", "prmm"),
    ],
)
def test_check_var_ranges_avail(data_tm5: GriddedData, dvar: str, var: str):
    data = data_tm5.copy()
    data.var_name = dvar
    check_var_ranges_avail(data, var)


def test_check_var_ranges_avail_error(data_tm5: GriddedData):
    with pytest.raises(VariableDefinitionError) as e:
        check_var_ranges_avail(data_tm5, "bla")
    assert str(e.value) == "Error (VarCollection): input variable bla is not supported"


def test__check_statistics_periods():
    assert _check_statistics_periods(["2010-2010", "2005"]) == ["2010-2010", "2005"]


@pytest.mark.parametrize(
    "periods,error",
    [
        pytest.param(
            42,
            "statistics_periods needs to be a list",
            id="not list",
        ),
        pytest.param(
            [42],
            "All periods need to be strings",
            id="not list[str]",
        ),
        pytest.param(
            ["2010-2010-20"],
            "Invalid value for period (2010-2010-20), can be either single years or period of years (e.g. 2000-2010).",
            id="wrong period",
        ),
    ],
)
def test__check_statistics_periods_error(periods: list[str], error: str):
    with pytest.raises(ValueError) as e:
        _check_statistics_periods(periods)
    assert str(e.value) == error


@pytest.mark.parametrize(
    "period,result",
    [
        ("2005", slice("2005", "2005")),
        ("2005-2019", slice("2005", "2019")),
    ],
)
def test__period_str_to_timeslice(period: str, result: slice):
    assert _period_str_to_timeslice(period) == result


def test__period_str_to_timeslice_error():
    with pytest.raises(ValueError) as e:
        _period_str_to_timeslice("2005-2019-2000")
    assert str(e.value) == "2005-2019-2000"


@pytest.mark.parametrize(
    "periods,result",
    [
        (["2005", "2000"], (2000, 2005)),
        (["2005", "2000", "1999-2021"], (1999, 2021)),
    ],
)
def test__get_min_max_year_periods(periods: list[str], result: tuple[int, int]):
    assert _get_min_max_year_periods(periods) == result


def test__get_min_max_year_periods_error():
    with pytest.raises(ValueError) as e:
        _get_min_max_year_periods(["2005-2004-23", "2000", "1999-2021"])
    assert str(e.value) == "2005-2004-23"


@pytest.mark.parametrize("cfg", ["cfgexp1"])
def test_make_dummy_model(eval_config: dict):
    cfg = EvalSetup(**eval_config)
    assert cfg.obs_cfg["AERONET-Sun"]
    model_id = make_dummy_model(["AERONET-Sun"], cfg)
    assert model_id == "dummy_model"



import sys

if sys.version_info >= (3, 10):  # pragma: no cover
    from importlib import metadata
else:  # pragma: no cover
    import importlib_metadata as metadata


def test_gridded():
    names = {ep.name for ep in metadata.entry_points(group="pyaerocom.gridded")}
    assert names, "no entry points found"
    assert names == {"ReadMscwCtm"}


def test_ungridded():
    names = {ep.name for ep in metadata.entry_points(group="pyaerocom.ungridded")}
    assert names, "no entry points found"
    assert names == {"ReadGAW", "ReadGhost", "ReadMEP", "ReadICOS", "ReadICPForest"}



from pathlib import Path

import numpy as np
import pytest

from pyaerocom.io.read_aeronet_sunv2 import ReadAeronetSunV2
from tests.conftest import TEST_RTOL, lustre_unavail


@lustre_unavail
def test_load_berlin_AeroSunV2L2D():
    reader = ReadAeronetSunV2()
    files = reader.find_in_file_list("*Berlin*")
    assert len(files) == 1
    assert Path(files[0]).name == "920801_180519_Berlin_FUB.lev20"
    data = reader.read_file(files[0], vars_to_retrieve=["od550aer"])

    test_vars = ["od440aer", "od500aer", "od550aer", "ang4487aer"]

    assert all(x in data for x in test_vars)

    # more than 100 timestamps
    assert all(len(data[x]) > 100 for x in test_vars)

    assert isinstance(data["dtime"][0], np.datetime64)
    assert data["dtime"][0] == np.datetime64("2014-07-06T00:00:00")

    first_vals = [data[var][0] for var in test_vars]
    nominal = [0.229427, 0.18302, 0.151227, 2.002052]
    assert first_vals == pytest.approx(nominal, rel=TEST_RTOL)
